from tempfile import mkstemp

from pygame.image import load, save, tostring, fromstring
from pygame.draw import lines
from pygame.transform import scale
from pygame import Surface

class Image:
	def __init__(self, file=None, size=None, surface=None, image=None, depth=None):
		if file:
			self.surface = load(file)
		elif size:
			self.surface = Surface(size, depth=depth)
		elif surface:
			self.surface = surface
		elif image:
			self.surface = image.surface.copy()
		else:
			self.surface = None
	
	def ToString(self):
		return tostring(self.surface, "RGBA")
	
	def FromString(self, s, size):
		self.surface = fromstring(s, size, "RGBA", flipped=False)
		return self
	
	def Scale(self, newsize, dest=None):
		if dest and dest.surface.get_size() == self.surface.get_size():
			newsurface = dest.surface
			self.surface = scale(self.surface, newsize, newsurface)
		else:
			self.surface = scale(self.surface, newsize)
		return self
	
	def Size(self):
		return self.surface.get_size()
	
	def SubImage(self, rectangle, copy=False):
		"""subimage returns a window into an image. If you want to create an entirely new surface, pass copy=True"""
		if copy:
			return self.__class__(surface=self.surface.subsurface(rectangle).copy())
		return self.__class__(surface=self.surface.subsurface(rectangle))
	
	def Copy(self):
		return self.__class__(surface=self.surface)
	
	def RemoveAlpha(self,bgcolor=(255,0,255)):
		"""This makes clear pixels paste in and destroy the destination pixel, rather than actually be transparent"""
		# this sub contains alpha. So lets fill those areas with the background colour
		surf = self.surface.copy()
		surf.set_colorkey((255,255,0))			# so the fill with background is not interpreted as "fill with transparet" TODO: make this more generic
		surf.fill(bgcolor)
		surf.blit(self.surface.convert(surf),(0,0))
		self.surface=surf
	
	def Pixel(self, pos, color=None):
		if color:
			self.surface.set_at(pos, color)
			return color
		else:
			return self.surface.get_at(pos)
	
	def TransparentColor(self, color=None):
		if color == -1:
			self.surface.set_colorkey(None)
		elif color != None:
			self.surface.set_colorkey(color)
		return self.surface.get_colorkey()
		
	def Line(self, points, color, closed=False):
		return lines(self.surface, color, closed, points, 1)
	
	def Palette(self, palette=None):
		if palette:
			self.surface.set_palette(palette)
		return self.surface.get_palette()
	
	def Alpha(self, alpha=None):
		if not alpha is None:
			self.surface.set_alpha(alpha)
		return self.surface.get_alpha()
	
	def Save(self, filename):
		save(self.surface, filename)
	
	def Fill(self, color):
		self.surface.fill(color)
	
	def FloodFill(self, pos, newColor):
		"""
			http://www.student.kuleuven.be/~m0216922/CG/floodfill.html
			
			The scanline floodfill algorithm using stack
			void floodFillScanlineStack(int x, int y, int newColor, int oldColor)
		"""
		STACKSIZE = 1048576
		s = []
		w = self.surface.get_width()
		h = self.surface.get_height()
		oldColor = self.Pixel(pos)
		bigsize = 0
		
		if oldColor != tuple(newColor):
			s.append(pos[:])
			y1 = 0
			spanLeft = 0
			spanRight = 0
			
			while len(s) and len(s) <= STACKSIZE:
				x, y = s.pop(0)
				y1 = y
				while (y1 >= 0 and self.Pixel((x, y1)) == oldColor):
					y1 -= 1
				y1 += 1
				spanLeft = 0
				spanRight = 0
				while y1 < h and self.Pixel((x, y1)) == oldColor:
					self.Pixel((x, y1), newColor)
					if not spanLeft and x > 0 and self.Pixel((x - 1, y1)) == oldColor:
						s.append((x - 1, y1))
						spanLeft = 1
					elif spanLeft and x > 0 and self.Pixel((x - 1, y1)) != oldColor:
						spanLeft = 0
					if not spanRight and x < w - 1 and self.Pixel((x + 1, y1)) == oldColor:
						s.append((x + 1, y1))
						spanRight = 1
					elif spanRight and x < w - 1 and self.Pixel((x + 1, y1)) != oldColor:
						spanRight = 0
					y1 += 1
			
			# this should really never happen on a 1024x1024 image
			if len(s) > STACKSIZE:
				raise Exception("Fill stack smashed!")
	
	def Blit(self, image, pos, size=None):
		"""paste an image into another image"""
		self.surface.blit(image.surface,pos)
		
		
